import tornado.template

from pypugjs import Compiler as _Compiler
from pypugjs.exceptions import CurrentlyNotSupported
from pypugjs.runtime import attrs, escape, iteration
from pypugjs.utils import process

ATTRS_FUNC = '__pypugjs_attrs'
ESCAPE_FUNC = '__pypugjs_escape'
ITER_FUNC = '__pypugjs_iter'


class Compiler(_Compiler):
    def visitCodeBlock(self, block):
        self.buffer('{%% block %s %%}' % block.name)
        if block.mode == 'append':
            self.buffer('{% raw super() %}')
        self.visitBlock(block)
        if block.mode == 'prepend':
            self.buffer('{% raw super() %}')
        self.buffer('{% end %}')

    def interpolate(self, text, escape=True):
        return self._interpolate(
            text, lambda x: '{%% raw %s(%s) %%}' % (ESCAPE_FUNC, x)
        )

    def visitMixin(self, mixin):
        raise CurrentlyNotSupported('mixin')

    def visitAssignment(self, assignment):
        self.buffer('{%% set %s = %s %%}' % (assignment.name, assignment.val))

    def visitCode(self, code):
        if code.buffer:
            val = code.val.lstrip()
            val = self.var_processor(val)
            if code.escape:
                self.buf.append('{%% raw %s(%s) %%}' % (ESCAPE_FUNC, val))
            else:
                self.buf.append('{%% raw %s %%}' % val)
        else:
            self.buf.append('{%% %s %%}' % code.val)

        if code.block:
            # if not code.buffer: self.buf.append('{')
            self.visit(code.block)
            # if not code.buffer: self.buf.append('}')

            if not code.buffer:
                codeTag = code.val.strip().split(' ', 1)[0]
                if codeTag in self.auto_close_code:
                    self.buf.append('{%% end%s %%}' % codeTag)

    def visitEach(self, each):
        self.buf.append(
            '{%% for %s in %s(%s,%s) %%}'
            % (','.join(each.keys), ITER_FUNC, each.obj, len(each.keys))
        )
        self.visit(each.block)
        self.buf.append('{% end %}')

    def visitConditional(self, conditional):
        TYPE_CODE = {
            'if': lambda x: 'if %s' % x,
            'unless': lambda x: 'if not %s' % x,
            'elif': lambda x: 'elif %s' % x,
            'else': lambda x: 'else',
        }
        self.buf.append(
            '{%% %s %%}' % TYPE_CODE[conditional.type](conditional.sentence)
        )
        if conditional.block:
            self.visit(conditional.block)
            for next in conditional.next:
                self.visitConditional(next)
        if conditional.type in ['if', 'unless']:
            self.buf.append('{% end %}')

    def attributes(self, attrs):
        return "{%% raw %s(%s) %%}" % (ATTRS_FUNC, attrs)


class Template(tornado.template.Template):
    def __init__(self, template_string, name="<string>", *args, **kwargs):
        is_pugjs = name.endswith(".pug")
        if is_pugjs:
            template_string = process(template_string, filename=name, compiler=Compiler)

        super(Template, self).__init__(template_string, name, *args, **kwargs)
        if is_pugjs:
            self.namespace.update(
                {ATTRS_FUNC: attrs, ESCAPE_FUNC: escape, ITER_FUNC: iteration}
            )


# Patch tornado template engine for preprocess PugJS templates
def patch_tornado():
    tornado.template.Template = Template



import codecs
import os
import sys
from optparse import OptionParser

from pypugjs.utils import process


def convert_file():
    support_compilers_list = [
        'html',
        'django',
        'jinja',
        'underscore',
        'mako',
        'tornado',
    ]

    usage = "usage: %prog [options] [file [output]]"
    parser = OptionParser(usage)
    parser.add_option(
        "-o", "--output", dest="output", help="Write output to FILE", metavar="FILE"
    )
    parser.add_option(
        "-c",
        "--compiler",
        dest="compiler",
        choices=support_compilers_list,
        default='html',
        help="Compiler to use (default: html)",
    )
    parser.add_option(
        "-e",
        "--ext",
        dest="extension",
        help="Set import/extends default file extension",
        metavar="FILE",
    )

    options, args = parser.parse_args()

    compiler_name = options.compiler

    # Compiler jetzt dynamisch laden
    try:
        compiler_module = __import__(
            f'pypugjs.ext.{compiler_name}', fromlist=['pypugjs']
        )
        compiler_class = compiler_module.Compiler
    except ImportError:
        sys.exit(
            f"Compiler '{compiler_name}' not available. Please install it, or use one of: {', '.join(support_compilers_list)}"
        )

    file_output = options.output or (args[1] if len(args) > 1 else None)

    if options.extension:
        extension = '.%s' % options.extension
    elif options.output:
        extension = os.path.splitext(options.output)[1]
    else:
        extension = None

    import six

    if len(args) >= 1:
        template = codecs.open(args[0], 'r', encoding='utf-8').read()
    elif six.PY3:
        template = sys.stdin.read()
    else:
        template = codecs.getreader('utf-8')(sys.stdin).read()

    output = process(
        template,
        compiler=compiler_class,
        staticAttrs=True,
        extension=extension,
    )

    if file_output:
        with codecs.open(file_output, 'w', encoding='utf-8') as outfile:
            outfile.write(output)
    elif six.PY3:
        sys.stdout.write(output)
    else:
        codecs.getwriter('utf-8')(sys.stdout).write(output)


if __name__ == '__main__':
    convert_file()



from itertools import count

import six

from pypugjs import Compiler as _Compiler


def process_param(key, value, terse=False):
    # Always render boolean True as presence-only attribute
    if value is True:
        return key
    if terse and (key == value):
        return key
    if isinstance(value, six.binary_type):
        value = value.decode('utf8')
    return '''%s="%s"''' % (key, value)


class Compiler(_Compiler):
    def __init__(self, *args, **kws):
        _Compiler.__init__(self, *args, **kws)
        self._i = count()

    def visitAssignment(self, assignment):
        self.buffer('<%% var %s = %s; %%>' % (assignment.name, assignment.val))

    def visitCode(self, code):
        if code.buffer:
            val = code.val.lstrip()
            self.buf.append('<%%%s %s %%>' % ('=' if code.escape else '-', val))
        else:
            self.buf.append('<%% %s' % code.val)  # for loop

        if code.block:
            self.buf.append(' { %>')  # for loop
            # if not code.buffer: self.buf.append('{')
            self.visit(code.block)
            # if not code.buffer: self.buf.append('}')

            if not code.buffer:
                codeTag = code.val.strip().split(' ', 1)[0]
                if codeTag in self.auto_close_code:
                    self.buf.append('<% } %>')
        elif not code.buffer:
            self.buf.append('; %>')  # for loop

    def visitEach(self, each):
        # self.buf.append('{%% for %s in %s %%}'%(','.join(each.keys),each.obj))
        __i = self._i.next()
        self.buf.append(
            '<%% for (_i_%s = 0, _len_%s = %s.length; _i_%s < _len_%s; _i_%s++) '
            '{ ' % (__i, __i, each.obj, __i, __i, __i)
        )
        if len(each.keys) > 1:
            for i, k in enumerate(each.keys):
                self.buf.append('%s = %s[_i_%s][%s];' % (k, each.obj, __i, i))
        else:
            for k in each.keys:
                self.buf.append('%s = %s[_i_%s];' % (k, each.obj, __i))
        self.buf.append(' %>')
        self.visit(each.block)
        self.buf.append('<% } %>')

    def _do_eval(self, value):
        if isinstance(value, six.string_types):
            value = value.encode('utf-8')
        try:
            value = eval(value, {}, {})
        except Exception:
            return "<%%= %s %%>" % value
        return value

    def _get_value(self, attr):
        value = attr['val']
        if attr['static']:
            return attr['val']
        if isinstance(value, six.string_types):
            return self._do_eval(value)
        else:
            return attr['name']

    def visitAttributes(self, attrs):
        classes = []
        params = []
        for attr in attrs:
            if attr['name'] == 'class':
                value = self._get_value(attr)
                if isinstance(value, list):
                    classes.extend(value)
                else:
                    classes.append(value)
            else:
                value = self._get_value(attr)
                if (value is not None) and (value is not False):
                    params.append((attr['name'], value))
        if classes:
            classes = [six.text_type(c) for c in classes]
            params.append(('class', " ".join(classes)))
        if params:
            self.buf.append(
                " " + " ".join([process_param(k, v, self.terse) for (k, v) in params])
            )

    def visitConditional(self, conditional):
        TYPE_CODE = {
            'if': lambda x: 'if (%s)' % x,
            'unless': lambda x: 'if (!%s)' % x,
            'elif': lambda x: '} else if (%s)' % x,
            'else': lambda x: '} else',
        }
        self.buf.append(
            '\n<%% %s { %%>' % TYPE_CODE[conditional.type](conditional.sentence)
        )
        if conditional.block:
            self.visit(conditional.block)
            for next in conditional.next:
                self.visitConditional(next)
        if conditional.type in ['if', 'unless']:
            self.buf.append('\n<% } %>\n')

    def interpolate(self, text, escape=True):
        return self._interpolate(text, lambda x: '<%%= %s %%>' % x)



import re
import os
import six


class Compiler(object):
    RE_INTERPOLATE = re.compile(r'(\\)?([#!]){(.*?)}')
    doctypes = {
        '5': '<!DOCTYPE html>',
        'xml': '<?xml version="1.0" encoding="utf-8" ?>',
        'default': '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" '
        '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',
        'transitional': '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" '
        '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',
        'strict': '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" '
        '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">',
        'frameset': '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" '
        '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd">',
        '1.1': '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" '
        '"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">',
        'basic': '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML Basic 1.1//EN" '
        '"http://www.w3.org/TR/xhtml-basic/xhtml-basic11.dtd">',
        'mobile': '<!DOCTYPE html PUBLIC "-//WAPFORUM//DTD XHTML Mobile 1.2//EN" '
        '"http://www.openmobilealliance.org/tech/DTD/xhtml-mobile12.dtd">',
    }
    inline_tags = [
        'a',
        'abbr',
        'acronym',
        'b',
        'br',
        'code',
        'em',
        'font',
        'i',
        'img',
        'ins',
        'kbd',
        'map',
        'samp',
        'small',
        'span',
        'strong',
        'sub',
        'sup',
        'textarea',
    ]
    self_closing = ['meta', 'img', 'link', 'input', 'area', 'base', 'col', 'br', 'hr']
    auto_close_code = [
        'if',
        'for',
        'block',
        'filter',
        'autoescape',
        'with',
        'trans',
        'spaceless',
        'comment',
        'cache',
        'macro',
        'localize',
        'compress',
        'ra,',
    ]
    filters = {}

    def __init__(self, node, **options):
        self.options = options
        self.node = node
        self.hasCompiledDoctype = False
        self.hasCompiledTag = False
        self.pp = options.get('pretty', True)
        self.debug = options.get('compileDebug', False) is not False
        self.filters.update(options.get('filters', {}))
        self.doctypes.update(options.get('doctypes', {}))
        # self.var_processor = options.get('var_processor', lambda x: x)
        self.self_closing.extend(options.get('self_closing', []))
        self.auto_close_code.extend(options.get('auto_close_code', []))
        self.inline_tags.extend(options.get('inline_tags', []))
        self.useRuntime = options.get('useRuntime', True)
        self.extension = options.get('extension', None) or '.pug'
        self.indents = 0
        self.doctype = None
        self.terse = False
        self.xml = False
        self.mixing = 0
        self.variable_start_string = options.get("variable_start_string", "{{")
        self.variable_end_string = options.get("variable_end_string", "}}")
        if 'doctype' in self.options:
            self.setDoctype(options['doctype'])
        self.instring = False

    def var_processor(self, var):
        if isinstance(var, six.string_types) and var.startswith('_ '):
            var = '_("%s")' % var[2:]
        return var

    def compile_top(self):
        return ''

    def compile(self):
        self.buf = [self.compile_top()]
        self.lastBufferedIdx = -1
        self.visit(self.node)
        compiled = u''.join(self.buf)
        if isinstance(compiled, six.binary_type):
            compiled = six.text_type(compiled, 'utf8')
        return compiled

    def setDoctype(self, name):
        self.doctype = self.doctypes.get(name or 'default', '<!DOCTYPE %s>' % name)
        self.terse = name in ['5', 'html']
        self.xml = self.doctype.startswith('<?xml')

    def buffer(self, str):
        if self.lastBufferedIdx == len(self.buf):
            self.lastBuffered += str
            self.buf[self.lastBufferedIdx - 1] = self.lastBuffered
        else:
            self.buf.append(str)
            self.lastBuffered = str
            self.lastBufferedIdx = len(self.buf)

    def visit(self, node, *args, **kwargs):
        # debug = self.debug
        # if debug:
        #     self.buf.append('__pugjs.unshift({ lineno: %d, filename: %s });' % (
        #         node.line, ('"%s"' % node.filename) if node.filename else '__pugjs[0].filename'))
        #
        # if node.debug==False and self.debug:
        #     self.buf.pop()
        #     self.buf.pop()

        self.visitNode(node, *args, **kwargs)
        # if debug: self.buf.append('__pugjs.shift();')

    def visitNode(self, node, *args, **kwargs):
        name = node.__class__.__name__
        if self.instring and name != 'Tag':
            self.buffer('\n')
            self.instring = False
        return getattr(self, 'visit%s' % name)(node, *args, **kwargs)

    def visitLiteral(self, node):
        self.buffer(node.str)

    def visitBlock(self, block):
        for node in block.nodes:
            self.visit(node)

    def visitCodeBlock(self, block):
        self.buffer('{%% block %s %%}' % block.name)
        if block.mode == 'prepend':
            self.buffer(
                '%ssuper()%s' % (self.variable_start_string, self.variable_end_string)
            )
        self.visitBlock(block)
        if block.mode == 'append':
            self.buffer(
                '%ssuper()%s' % (self.variable_start_string, self.variable_end_string)
            )
        self.buffer('{% endblock %}')

    def visitDoctype(self, doctype=None):
        if doctype and (doctype.val or not self.doctype):
            self.setDoctype(doctype.val or 'default')

        if self.doctype:
            self.buffer(self.doctype)
        self.hasCompiledDoctype = True

    def visitMixin(self, mixin):
        if mixin.block:
            self.buffer('{%% macro %s(%s) %%}' % (mixin.name, mixin.args))
            self.visitBlock(mixin.block)
            self.buffer('{% endmacro %}')
        else:
            self.buffer(
                '%s%s(%s)%s'
                % (
                    self.variable_start_string,
                    mixin.name,
                    mixin.args,
                    self.variable_end_string,
                )
            )

    def visitTag(self, tag):
        self.indents += 1
        name = tag.name
        if not self.hasCompiledTag:
            if not self.hasCompiledDoctype and 'html' == name:
                self.visitDoctype()
            self.hasCompiledTag = True

        if self.pp and name not in self.inline_tags and not tag.inline:
            self.buffer('\n' + '  ' * (self.indents - 1))
        if name in self.inline_tags or tag.inline:
            self.instring = False

        closed = name in self.self_closing and not self.xml
        if tag.text:
            t = tag.text.nodes[0]
            if t.startswith(u'/'):
                if len(t) > 1:
                    raise Exception(
                        '%s is self closing and should not have content.' % name
                    )
                closed = True

        if tag.buffer:
            self.buffer('<' + self.interpolate(name))
        else:
            self.buffer('<%s' % name)
        self.visitAttributes(tag.attrs)
        self.buffer('/>' if not self.terse and closed else '>')

        if not closed:
            if tag.code:
                self.visitCode(tag.code)
            if tag.text:
                text_val = tag.text.nodes[0]
                if text_val.startswith(' '):
                    text_val = text_val[1:]
                self.buffer(self.interpolate(text_val))
            self.escape = 'pre' == tag.name
            # empirically check if we only contain text
            textOnly = tag.textOnly or not bool(len(tag.block.nodes))
            self.instring = False
            self.visit(tag.block)

            if self.pp and name not in self.inline_tags and not textOnly:
                self.buffer('\n' + '  ' * (self.indents - 1))

            if tag.buffer:
                self.buffer('</' + self.interpolate(name) + '>')
            else:
                self.buffer('</%s>' % name)
        self.indents -= 1

    def visitFilter(self, filter):
        if filter.name not in self.filters:
            if filter.isASTFilter:
                raise Exception('unknown ast filter "%s"' % filter.name)
            else:
                raise Exception('unknown filter "%s"' % filter.name)

        fn = self.filters.get(filter.name)
        if filter.isASTFilter:
            self.buf.append(fn(filter.block, self, filter.attrs))
        else:
            text = ''.join(filter.block.nodes)
            text = self.interpolate(text)
            filter.attrs = filter.attrs or {}
            filter.attrs['filename'] = self.options.get('filename', None)
            self.buffer(fn(text, filter.attrs))

    def html_escape(self, s):
        return (s
                .replace('&', '&amp;')
                .replace('<', '&lt;')
                .replace('>', '&gt;')
                )

    def _interpolate(self, attr, repl):
        return self.RE_INTERPOLATE.sub(lambda matchobj: repl(matchobj.group(3)), attr)

    def interpolate(self, text, escape=None):
        def repl(matchobj):
            if escape is None:
                if matchobj.group(2) == '!':
                    filter_string = ''
                else:
                    filter_string = '|escape'
            else:
                if escape is True:
                    filter_string = '|escape'
                else:
                    filter_string = ''

            matchvar = matchobj.group(3)
            if (matchvar[0] in ['"', "'"]
                    and matchvar[-1] == matchvar[0]
                    and filter_string == '|escape'):
                # Django doesn't correctly escape strings. To stay consistent
                # with Pug, escape these now.
                return self.html_escape(matchvar[1:-1])

            return (
                self.variable_start_string
                + matchvar
                + filter_string
                + self.variable_end_string
            )

        return self.RE_INTERPOLATE.sub(repl, text)

    def visitText(self, text):
        text = ''.join(text.nodes)
        text = self.interpolate(text)
        self.buffer(text)
        if self.pp:
            self.buffer('\n')

    def visitString(self, text):
        instring = not text.inline
        text = ''.join(text.nodes)
        text = self.interpolate(text)
        self.buffer(text)
        self.instring = instring

    def visitComment(self, comment):
        if not comment.buffer:
            return
        if self.pp:
            self.buffer('\n' + '  ' * (self.indents))
        self.buffer('<!--%s-->' % comment.val)

    def visitAssignment(self, assignment):
        self.buffer('{%% set %s = %s %%}' % (assignment.name, assignment.val))

    def format_path(self, path):
        has_extension = '.' in os.path.basename(path)
        if not has_extension:
            path += self.extension
        return path

    def visitExtends(self, node):
        path = self.format_path(node.path)
        self.buffer('{%% extends "%s" %%}' % (path))

    def visitInclude(self, node):
        path = self.format_path(node.path)
        self.buffer('{%% include "%s" %%}' % (path))

    def visitBlockComment(self, comment):
        if not comment.buffer:
            return
        isConditional = comment.val.strip().startswith('if')
        self.buffer(
            '<!--[%s]>' % comment.val.strip()
            if isConditional
            else '<!--%s' % comment.val
        )
        self.visit(comment.block)
        self.buffer('<![endif]-->' if isConditional else '-->')

    def visitConditional(self, conditional):
        TYPE_CODE = {
            'if': lambda x: 'if %s' % x,
            'unless': lambda x: 'if not %s' % x,
            'elif': lambda x: 'elif %s' % x,
            'else': lambda x: 'else',
        }
        self.buf.append(
            '{%% %s %%}' % TYPE_CODE[conditional.type](conditional.sentence)
        )
        if conditional.block:
            self.visit(conditional.block)
            for next in conditional.next:
                self.visitConditional(next)
        if conditional.type in ['if', 'unless']:
            self.buf.append('{% endif %}')

    def visitVar(self, var, escape=False):
        var = self.var_processor(var)
        return '%s%s%s%s' % (
            self.variable_start_string,
            var,
            '|escape' if escape else '',
            self.variable_end_string,
        )

    def visitCode(self, code):
        if code.buffer:
            val = code.val.lstrip()

            self.buf.append(self.visitVar(val, code.escape))
        else:
            self.buf.append('{%% %s %%}' % code.val)

        if code.block:
            # if not code.buffer: self.buf.append('{')
            self.visit(code.block)
            # if not code.buffer: self.buf.append('}')

            if not code.buffer:
                code_tag = code.val.strip().split(' ', 1)[0]
                if code_tag in self.auto_close_code:
                    self.buf.append('{%% end%s %%}' % code_tag)

    def visitEach(self, each):
        self.buf.append(
            '{%% for %s in %s|__pypugjs_iter:%d %%}'
            % (','.join(each.keys), each.obj, len(each.keys))
        )
        self.visit(each.block)
        self.buf.append('{% endfor %}')

    def attributes(self, attrs):
        return "%s__pypugjs_attrs(%s)%s" % (
            self.variable_start_string,
            attrs,
            self.variable_end_string,
        )

    def visitDynamicAttributes(self, attrs):
        buf, classes, params = [], [], {}
        terse = 'terse=True' if self.terse else ''
        for attr in attrs:
            if attr['name'] == 'class':
                classes.append('(%s)' % attr['val'])
            else:
                pair = "('%s',(%s))" % (attr['name'], attr['val'])
                buf.append(pair)

        if classes:
            classes = " , ".join(classes)
            buf.append("('class', (%s))" % classes)

        buf = ', '.join(buf)
        if self.terse:
            params['terse'] = 'True'
        if buf:
            params['attrs'] = '[%s]' % buf
        param_string = ', '.join(['%s=%s' % (n, v) for n, v in six.iteritems(params)])
        if buf or terse:
            self.buf.append(self.attributes(param_string))

    def visitAttributes(self, attrs):
        temp_attrs = []
        for attr in attrs:
            if (not self.useRuntime and not attr['name'] == 'class') or attr['static']:
                if temp_attrs:
                    self.visitDynamicAttributes(temp_attrs)
                    temp_attrs = []
                n, v = attr['name'], attr['val']
                if isinstance(v, six.string_types):
                    if self.useRuntime or attr['static']:
                        self.buf.append(' %s=%s' % (n, v))
                    else:
                        self.buf.append(' %s="%s"' % (n, self.visitVar(v)))
                elif v is True:
                    # For boolean attributes, prefer presence form regardless of doctype
                    # (e.g., checked, selected, v-else)
                    self.buf.append(' %s' % (n,))
            else:
                temp_attrs.append(attr)

        if temp_attrs:
            self.visitDynamicAttributes(temp_attrs)

    @classmethod
    def register_filter(cls, name, f):
        cls.filters[name] = f

    @classmethod
    def register_autoclosecode(cls, name):
        cls.auto_close_code.append(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version("pypugjs", "__init__.py")


readme = open('README.rst').read()
history = open('docs/HISTORY.rst').read().replace('.. :changelog:', '')
url = 'https://github.com/kakulukia/pypugjs'

setup(
    name='pypugjs',
    version=version,
    description="PugJS syntax template adapter for Django, Jinja2, Mako and Tornado templates",
    long_description=readme + '\n\n' + history,
    author='Andy Grabow',
    author_email='andy@freilandkiwis.de',
    license='MIT',
    keywords=['pug', 'pugjs', 'template', 'converter'],
    url=url,
    download_url=url + '/tarball/' + version,
    packages=find_packages(),
    entry_points={'console_scripts': ['pypugjs = pypugjs.convert:convert_file']},
    install_requires=['six', 'charset_normalizer'],
    tests_require=[
        'pytest',
        'pytest-cov',
        'django',
        'jinja2',
        'tornado',
        'pyramid >= 1.5',
        'pyramid-mako',
        'mako',
        'icecream',
    ],
    # include_package_data=True,
    # zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 4.0',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from __future__ import absolute_import

import re
from collections import deque

import six


class Token:
    def __init__(self, **kwds):
        self.buffer = None
        self.__dict__.update(kwds)

    def __str__(self):
        return self.__dict__.__str__()


def regexec(regex, input):
    matches = regex.match(input)
    if matches:
        return (input[matches.start() : matches.end()],) + matches.groups()
    return None


def detect_closing_bracket(string):
    count = 0
    pos = string.find('[')
    while True:
        if string[pos] == '[':
            count += 1
        if string[pos] == ']':
            count -= 1
        pos += 1
        if count == 0:
            return pos


def replace_string_brackets(splitted_string):
    sval_replaced = []
    old_delim = None
    for i in splitted_string:
        if old_delim is None:
            sval_replaced.append(i)
            if i in ('"', "'"):
                old_delim = i
            continue

        if i in ('"', "'"):
            if i == old_delim:
                old_delim = None
            sval_replaced.append(i)
            continue

        sval_replaced.append(re.sub(r'\[|\]', '*', i))
    return ''.join(sval_replaced)


class Lexer(object):
    RE_INPUT = re.compile(r'\r\n|\r')
    RE_COMMENT = re.compile(r'^ *\/\/(-)?([^\n]*)')
    RE_TAG = re.compile(r'^(\w[-:\w]*|#\{.*?\})')
    RE_DOT_BLOCK_START = re.compile(r'^\.\n')
    RE_FILTER = re.compile(r'^:(\w+)')
    RE_DOCTYPE = re.compile(r'^(?:!!!|doctype) *([^\n]+)?')
    RE_ID = re.compile(r'^#([\w-]+)')
    RE_CLASS = re.compile(r'^\.([\w-]+)')
    RE_STRING = re.compile(r'^(?:\| ?)([^\n]*)')
    RE_TEXT = re.compile(r'^([^\n]+)')
    RE_EXTENDS = re.compile(r'^extends? +([^\n]+)')
    RE_PREPEND = re.compile(r'^prepend +([^\n]+)')
    RE_APPEND = re.compile(r'^append +([^\n]+)')
    RE_BLOCK = re.compile(r'''^block(( +(?:(prepend|append) +)?([^\n]*))|\n)''')
    RE_YIELD = re.compile(r'^yield *')
    RE_INCLUDE = re.compile(r'^include +([^\n]+)')
    RE_ASSIGNMENT = re.compile(r'^(-\s+var\s+)?(\w+) += *([^;\n]+)( *;? *)')
    RE_MIXIN = re.compile(r'^mixin +([-\w]+)(?: *\((.*)\))?')
    RE_CALL = re.compile(r'^\+\s*([-.\w]+)(?: *\((.*)\))?')
    RE_CONDITIONAL = re.compile(r'^(?:- *)?(if|unless|else if|elif|else)\b([^\n]*)')
    RE_BLANK = re.compile(r'^\n *\n')
    # RE_WHILE = re.compile(r'^while +([^\n]+)')
    RE_EACH = re.compile(r'^(?:- *)?(?:each|for) +([\w, ]+) +in +([^\n]+)')
    RE_CODE = re.compile(r'^(!?=|-)([^\n]+)')
    RE_ATTR_INTERPOLATE = re.compile(r'#\{([^}]+)\}')
    RE_ATTR_PARSE = re.compile(r'''^['"]|['"]$''')
    RE_INDENT_TABS = re.compile(r'^\n(\t*) *')
    RE_INDENT_SPACES = re.compile(r'^\n( *)')
    RE_COLON = re.compile(r'^: *')
    RE_INLINE = re.compile(r'(?<!\\)#\[')
    RE_INLINE_ESCAPE = re.compile(r'\\#\[')
    STRING_SPLITS = re.compile(r'([\'"])(.*?)(?<!\\)(\1)')

    def __init__(self, string, **options):
        if isinstance(string, six.binary_type):
            string = six.text_type(string, 'utf8')
        self.options = options
        self.input = self.RE_INPUT.sub('\n', string)
        self.colons = self.options.get('colons', False)
        self.deferredTokens = deque()
        self.lastIndents = 0
        self.lineno = 1
        self.stash = deque()
        self.indentStack = deque()
        self.indentRe = None
        self.pipeless = False
        self.isTextBlock = False

    def tok(self, type, val=None):
        return Token(
            type=type,
            line=self.lineno,
            val=val,
            inline_level=self.options.get('inline_level', 0),
        )

    def consume(self, len):
        self.input = self.input[len:]

    def scan(self, regexp, type):
        captures = regexec(regexp, self.input)
        # print regexp,type, self.input, captures
        if captures:
            # print captures
            self.consume(len(captures[0]))
            # print 'a',self.input
            if len(captures) == 1:
                return self.tok(type, None)
            return self.tok(type, captures[1])

    def defer(self, tok):
        self.deferredTokens.append(tok)

    def lookahead(self, n):
        # print self.stash
        fetch = n - len(self.stash)
        while True:
            fetch -= 1
            if not fetch >= 0:
                break
            self.stash.append(self.next())
        return self.stash[n - 1]

    def indexOfDelimiters(self, start, end):
        str, nstart, nend, pos = self.input, 0, 0, 0
        for i, s in enumerate(str):
            if start == s:
                nstart += 1
            elif end == s:
                nend += 1
                if nend == nstart:
                    pos = i
                    break
        return pos

    def stashed(self):
        # print self.stash
        return len(self.stash) and self.stash.popleft()

    def deferred(self):
        return len(self.deferredTokens) and self.deferredTokens.popleft()

    def eos(self):
        # print 'eos',bool(self.input)
        if self.input:
            return
        if self.indentStack:
            self.indentStack.popleft()
            return self.tok('outdent')
        else:
            return self.tok('eos')

    def consumeBlank(self):
        captures = regexec(self.RE_BLANK, self.input)
        if not captures:
            return

        self.lineno += 1
        self.consume(len(captures[0]) - 1)
        return captures

    def blank(self):
        if self.pipeless:
            return
        if self.consumeBlank():
            return self.next()

    def comment(self):
        captures = regexec(self.RE_COMMENT, self.input)
        if captures:
            self.consume(len(captures[0]))
            tok = self.tok('comment', captures[2])
            tok.buffer = '-' != captures[1]
            return tok

    def tag(self):
        captures = regexec(self.RE_TAG, self.input)
        # print self.input,captures,re.match('^(\w[-:\w]*)',self.input)
        if captures:
            self.consume(len(captures[0]))
            name = captures[1]
            if name.endswith(':'):
                name = name[:-1]
                tok = self.tok('tag', name)
                self.defer(self.tok(':'))
                while self.input[0] == ' ':
                    self.input = self.input[1:]
            else:
                tok = self.tok('tag', name)
            return tok

    def textBlockStart(self):
        captures = regexec(self.RE_DOT_BLOCK_START, self.input)
        if captures is None:
            return

        if len(self.indentStack) > 0:
            self.textBlockTagIndent = self.indentStack[0]
        else:
            self.textBlockTagIndent = 0

        self.consume(1)
        self.isTextBlock = True
        return self.textBlockContinue(isStart=True)

    def textBlockContinue(self, isStart=False):
        if not self.isTextBlock:
            return

        tokens = deque()
        while True:
            if self.consumeBlank():
                if not isStart:
                    tokens.append(self.tok('string', ''))
                continue

            eos = self.eos()
            if eos is not None:
                if isStart:
                    return eos
                tokens.append(eos)
                break

            nextIndent = self.captureIndent()
            if nextIndent is None or len(nextIndent[1]) <= self.textBlockTagIndent:
                self.isTextBlock = False
                if isStart:
                    return self.tok('newline')
                break

            padding = 0
            if not isStart and len(nextIndent[1]) > self.textBlockIndent:
                padding = len(nextIndent[1]) - self.textBlockIndent
                self.consume(1 + padding)
                self.input = '\n' + self.input

            indent = self.indent()
            if isStart:
                self.textBlockIndent = indent.val
                padding = 0

            itoks = self.scanInline(self.RE_TEXT, 'string')
            indentChar = self.indentRe == self.RE_INDENT_TABS and '\t' or ' '
            if itoks:
                itoks[0].val = (indentChar * padding) + itoks[0].val

            if isStart:
                for tok in itoks or []:
                    self.defer(tok)
                return indent

            tokens.extend(itoks)

        if not tokens:
            firstTok = None
        else:
            firstTok = tokens.popleft()
            while tokens:
                if tokens[-1].type == 'string' and not tokens[-1].val:
                    tokens.pop()
                    continue
                self.defer(tokens.popleft())

        self.isTextBlock = False
        return firstTok

    def filter(self):
        return self.scan(self.RE_FILTER, 'filter')

    def doctype(self):
        # print self.scan(self.RE_DOCTYPE, 'doctype')
        return self.scan(self.RE_DOCTYPE, 'doctype')

    def id(self):
        return self.scan(self.RE_ID, 'id')

    def className(self):
        return self.scan(self.RE_CLASS, 'class')

    def processInline(self, val):
        sval = self.STRING_SPLITS.split(val)
        sval_stripped = [i.strip() for i in sval]

        if sval_stripped.count('"') % 2 != 0 or sval_stripped.count("'") % 2 != 0:
            raise Exception(
                'Unbalanced quotes found inside inline PugJS at line %s.' % self.lineno
            )

        sval_replaced = replace_string_brackets(sval)
        start_inline = self.RE_INLINE.search(sval_replaced).start()

        try:
            closing = start_inline + detect_closing_bracket(
                sval_replaced[start_inline:]
            )
        except IndexError:
            raise Exception(
                'The end of the string was reached with no closing bracket found at line %s.'
                % self.lineno
            )

        textl = val[:start_inline]
        code = val[start_inline:closing][2:-1]
        textr = val[closing:]

        toks = deque()

        toks.append(self.tok('string', self.RE_INLINE_ESCAPE.sub('#[', textl)))

        ilexer = InlineLexer(code, inline_level=self.options.get('inline_level', 0) + 1)
        while True:
            tok = ilexer.advance()
            if tok.type == 'eos':
                break
            toks.append(tok)

        if self.RE_INLINE.search(textr):
            toks.extend(self.processInline(textr))
        else:
            toks.append(self.tok('string', self.RE_INLINE_ESCAPE.sub('#[', textr)))

        return toks

    def scanInline(self, regexp, type):
        ret = self.scan(regexp, type)
        if ret is None:
            return ret

        if self.RE_INLINE.search(ret.val):
            ret = self.processInline(ret.val)
            if ret:
                ret[0].val = ret[0].val.lstrip()
        else:
            ret.val = self.RE_INLINE_ESCAPE.sub('#[', ret.val)
            ret = deque([ret])
        return ret

    def scanInlineProcess(self, regexp, type_):
        toks = self.scanInline(regexp, type_)
        if not toks:
            return None

        firstTok = toks.popleft()
        for tok in toks:
            self.defer(tok)
        return firstTok

    def string(self):
        return self.scanInlineProcess(self.RE_STRING, 'string')

    def text(self):
        return self.scanInlineProcess(self.RE_TEXT, 'text')

    def extends(self):
        return self.scan(self.RE_EXTENDS, 'extends')

    def prepend(self):
        captures = regexec(self.RE_PREPEND, self.input)
        if captures:
            self.consume(len(captures[0]))
            mode, name = 'prepend', captures[1]
            tok = self.tok('block', name)
            tok.mode = mode
            return tok

    def append(self):
        captures = regexec(self.RE_APPEND, self.input)
        if captures:
            self.consume(len(captures[0]))
            mode, name = 'append', captures[1]
            tok = self.tok('block', name)
            tok.mode = mode
            return tok

    def block(self):
        captures = regexec(self.RE_BLOCK, self.input)
        if captures:
            self.consume(len(captures[0]))
            mode = captures[3] or 'replace'
            name = captures[4] or ''
            tok = self.tok('block', name)
            tok.mode = mode
            return tok

    def _yield(self):
        return self.scan(self.RE_YIELD, 'yield')

    def include(self):
        return self.scan(self.RE_INCLUDE, 'include')

    def assignment(self):
        captures = regexec(self.RE_ASSIGNMENT, self.input)
        if captures:
            self.consume(len(captures[0]))
            name, val = captures[2:4]
            tok = self.tok('assignment')
            tok.name = name
            tok.val = val
            return tok

    def mixin(self):
        captures = regexec(self.RE_MIXIN, self.input)
        if captures:
            self.consume(len(captures[0]))
            tok = self.tok('mixin', captures[1])
            tok.args = captures[2]
            return tok

    def call(self):
        captures = regexec(self.RE_CALL, self.input)
        if captures:
            self.consume(len(captures[0]))
            tok = self.tok('call', captures[1])
            tok.args = captures[2]
            return tok

    def conditional(self):
        captures = regexec(self.RE_CONDITIONAL, self.input)
        if captures:
            self.consume(len(captures[0]))
            type, sentence = captures[1:]
            tok = self.tok('conditional', type)
            tok.sentence = sentence
            return tok

    # def _while(self):
    #     captures = regexec(self.RE_WHILE,self.input)
    #     if captures:
    #         self.consume(len(captures[0]))
    #         return self.tok('code','while(%s)'%captures[1])

    def each(self):
        captures = regexec(self.RE_EACH, self.input)
        if captures:
            self.consume(len(captures[0]))
            tok = self.tok('each', None)
            tok.keys = [x.strip() for x in captures[1].split(',')]
            tok.code = captures[2]
            return tok

    def code(self):
        captures = regexec(self.RE_CODE, self.input)
        if captures:
            self.consume(len(captures[0]))
            flags, name = captures[1:]
            tok = self.tok('code', name)
            tok.escape = flags.startswith('=')
            # print captures
            tok.buffer = '=' in flags
            # print tok.buffer
            return tok

    def attrs(self):
        if '(' == self.input[0]:
            index = self.indexOfDelimiters('(', ')')
            string = self.input[1:index]
            tok = self.tok('attrs')
            colons = self.colons
            states = ['key']

            class Namespace:
                key = u''
                val = u''
                quote = u''
                literal = True

                def reset(self):
                    self.key = self.val = self.quote = u''
                    self.literal = True

                def __str__(self):
                    return dict(
                        key=self.key,
                        val=self.val,
                        quote=self.quote,
                        literal=self.literal,
                    ).__str__()

            ns = Namespace()

            def state():
                return states[-1]

            def interpolate(attr):
                attr, num = self.RE_ATTR_INTERPOLATE.subn(
                    lambda matchobj: '%s+"{}".format(%s)+%s'
                    % (ns.quote, matchobj.group(1), ns.quote),
                    attr,
                )
                return attr, (num > 0)

            self.consume(index + 1)
            from .utils import odict

            tok.attrs = odict()
            tok.static_attrs = set()
            str_nums = list(map(str, range(10)))

            # print '------'
            def parse(c):
                real = c
                if colons and ':' == c:
                    c = '='
                ns.literal = ns.literal and (state() not in ('object', 'array', 'expr'))
                # print ns, c, states
                if c in (',', '\n') or (
                    c == ' '
                    and state() == 'val'
                    and ns.val.strip()
                ):
                    s = state()
                    if s in ('expr', 'array', 'string', 'object'):
                        ns.val += c
                    else:
                        states.append('key')
                        ns.val = ns.val.strip()
                        ns.key = ns.key.strip()
                        if not ns.key:
                            return
                        # ns.literal = ns.quote
                        if not ns.literal:
                            if '!' == ns.key[-1]:
                                ns.literal = True
                                ns.key = ns.key[:-1]
                        ns.key = ns.key.strip("'\"")
                        if not ns.val:
                            tok.attrs[ns.key] = True
                        else:
                            val, is_interpolated = interpolate(ns.val)
                            if ns.key == 'class' and 'class' in tok.attrs:
                                tok.attrs['class'] = '[%s, %s]' % (
                                    tok.attrs['class'],
                                    val,
                                )
                            else:
                                tok.attrs[ns.key] = val
                            ns.literal = ns.literal and not is_interpolated
                        if ns.literal:
                            tok.static_attrs.add(ns.key)
                        ns.reset()
                elif '=' == c:
                    s = state()
                    if s == 'key char':
                        ns.key += real
                    elif s in ('val', 'expr', 'array', 'string', 'object'):
                        ns.val += real
                    else:
                        states.append('val')
                elif '(' == c:
                    if state() in ('val', 'expr'):
                        states.append('expr')
                    ns.val += c
                elif ')' == c:
                    if state() in ('val', 'expr'):
                        states.pop()
                    ns.val += c
                elif '{' == c:
                    if 'val' == state():
                        states.append('object')
                    ns.val += c
                elif '}' == c:
                    if 'object' == state():
                        states.pop()
                    ns.val += c
                elif '[' == c:
                    if 'val' == state():
                        states.append('array')
                    ns.val += c
                elif ']' == c:
                    if 'array' == state():
                        states.pop()
                    ns.val += c
                elif c in ('"', "'"):
                    s = state()
                    if 'key' == s:
                        states.append('key char')
                    elif 'key char' == s:
                        states.pop()
                    elif 'string' == s:
                        if c == ns.quote:
                            states.pop()
                        ns.val += c
                    else:
                        states.append('string')
                        ns.val += c
                        ns.quote = c
                elif '' == c:
                    pass
                else:
                    s = state()
                    ns.literal = ns.literal and (
                        s in ('key', 'string') or c in str_nums
                    )
                    # print c, s, ns.literal
                    if s in ('key', 'key char'):
                        ns.key += c
                    else:
                        ns.val += c

            for char in string:
                parse(char)

            parse(',')

            return tok

    def captureIndent(self):
        if self.indentRe:
            captures = regexec(self.indentRe, self.input)
        else:
            regex = self.RE_INDENT_TABS
            captures = regexec(regex, self.input)
            if captures and not captures[1]:
                regex = self.RE_INDENT_SPACES
                captures = regexec(regex, self.input)
            if captures and captures[1]:
                self.indentRe = regex
        return captures

    def indent(self):
        captures = self.captureIndent()

        if captures:
            indents = len(captures[1])
            self.lineno += 1
            self.consume(indents + 1)

            if not self.input:
                return self.tok('newline')
            if self.input[0] in (' ', '\t'):
                raise Exception(
                    'Invalid indentation, you can use tabs or spaces but not both'
                )

            if '\n' == self.input[0]:
                return self.tok('newline')

            if self.indentStack and indents < self.indentStack[0]:
                while self.indentStack and self.indentStack[0] > indents:
                    self.stash.append(self.tok('outdent'))
                    self.indentStack.popleft()
                tok = self.stash.pop()
            elif indents and (not self.indentStack or indents != self.indentStack[0]):
                self.indentStack.appendleft(indents)
                tok = self.tok('indent', indents)
            else:
                tok = self.tok('newline')

            return tok

    def pipelessText(self):
        if self.pipeless:
            if '\n' == self.input[0]:
                return
            i = self.input.find('\n')
            if -1 == i:
                i = len(self.input)
            str = self.input[:i]
            self.consume(len(str))
            return self.tok('text', str)

    def colon(self):
        return self.scan(self.RE_COLON, ':')

    def advance(self):
        return self.stashed() or self.next()

    def next(self):
        return (
            self.deferred()
            or self.textBlockContinue()
            or self.blank()
            or self.eos()
            or self.pipelessText()
            or self._yield()
            or self.doctype()
            or self.extends()
            or self.append()
            or self.prepend()
            or self.block()
            or self.include()
            or self.mixin()
            or self.call()
            or self.conditional()
            or self.each()
            or self.assignment()
            or self.tag()
            or self.textBlockStart()
            or self.filter()
            or self.code()
            or self.id()
            or self.className()
            or self.attrs()
            or self.indent()
            or self.comment()
            or self.colon()
            or self.string()
            or self.text()
        )


class InlineLexer(Lexer):
    def next(self):
        return (
            self.deferred()
            or self.blank()
            or self.eos()
            or self.pipelessText()
            or self.mixin()
            or self.call()
            or self.assignment()
            or self.tag()
            or self.code()
            or self.id()
            or self.className()
            or self.attrs()
            or self.colon()
            or self.string()
            or self.text()
        )



from pypugjs import Compiler as _Compiler
from pypugjs.utils import process

ATTRS_FUNC = '__pypugjs_attrs'
ITER_FUNC = '__pypugjs_iter'


class Compiler(_Compiler):
    useRuntime = True

    def compile_top(self):
        return (
            '# -*- coding: utf-8 -*-\n<%%! from pypugjs.runtime import attrs as %s, '
            'iteration as %s\nfrom mako.runtime import Undefined %%>'
            % (ATTRS_FUNC, ITER_FUNC)
        )

    def interpolate(self, text, escape=True):
        filt = '|h' if escape else ''
        return self._interpolate(text, lambda x: '${%s%s}' % (x, filt))

    def visitCodeBlock(self, block):
        if self.mixing > 0:
            self.buffer('${caller.body() if caller else ""}')
        else:
            self.buffer('<%%block name="%s">' % block.name)
            if block.mode == 'append':
                self.buffer('${parent.%s()}' % block.name)
            self.visitBlock(block)
            if block.mode == 'prepend':
                self.buffer('${parent.%s()}' % block.name)
            self.buffer('</%block>')

    def visitMixin(self, mixin):
        self.mixing += 1
        if not mixin.call:
            self.buffer('<%%def name="%s(%s)">' % (mixin.name, mixin.args))
            self.visitBlock(mixin.block)
            self.buffer('</%def>')
        elif mixin.block:
            self.buffer('<%%call expr="%s(%s)">' % (mixin.name, mixin.args))
            self.visitBlock(mixin.block)
            self.buffer('</%call>')
        else:
            self.buffer('${%s(%s)}' % (mixin.name, mixin.args))
        self.mixing -= 1

    def visitAssignment(self, assignment):
        self.buffer('<%% %s = %s %%>' % (assignment.name, assignment.val))

    def visitExtends(self, node):
        path = self.format_path(node.path)
        self.buffer('<%%inherit file="%s"/>' % path)

    def visitInclude(self, node):
        path = self.format_path(node.path)
        self.buffer('<%%include file="%s"/>' % path)
        self.buffer('<%%namespace file="%s" import="*"/>' % path)

    def visitConditional(self, conditional):
        TYPE_CODE = {
            'if': lambda x: 'if %s' % x,
            'unless': lambda x: 'if not %s' % x,
            'elif': lambda x: 'elif %s' % x,
            'else': lambda x: 'else',
        }
        self.buf.append(
            '\\\n%% %s:\n' % TYPE_CODE[conditional.type](conditional.sentence)
        )
        if conditional.block:
            self.visit(conditional.block)
            for next in conditional.next:
                self.visitConditional(next)
        if conditional.type in ['if', 'unless']:
            self.buf.append('\\\n% endif\n')

    def visitVar(self, var, escape=False):
        return '${%s%s}' % (var, '| h' if escape else '| n')

    def visitCode(self, code):
        if code.buffer:
            val = code.val.lstrip()
            val = self.var_processor(val)
            self.buf.append(self.visitVar(val, code.escape))
        else:
            self.buf.append('<%% %s %%>' % code.val)

        if code.block:
            # if not code.buffer: self.buf.append('{')
            self.visit(code.block)
            # if not code.buffer: self.buf.append('}')

            if not code.buffer:
                codeTag = code.val.strip().split(' ', 1)[0]
                if codeTag in self.auto_close_code:
                    self.buf.append('</%%%s>' % codeTag)

    def visitEach(self, each):
        self.buf.append(
            '\\\n%% for %s in %s(%s,%d):\n'
            % (','.join(each.keys), ITER_FUNC, each.obj, len(each.keys))
        )
        self.visit(each.block)
        self.buf.append('\\\n% endfor\n')

    def attributes(self, attrs):
        return "${%s(%s, undefined=Undefined) | n}" % (ATTRS_FUNC, attrs)


def preprocessor(source):
    return process(source, compiler=Compiler)



from __future__ import absolute_import

__version__ = '5.13.0'

from .compiler import Compiler  # noqa
from .ext import html
from .filters import register_filter  # noqa
from .parser import Parser  # noqa
from .utils import process  # noqa


def simple_convert(template):
    return html.process_pugjs(template)



from __future__ import absolute_import

import os
import re

from django.conf import settings
from django.template import TemplateDoesNotExist
from django.template.loaders import cached

from pypugjs.utils import process
from .compiler import Compiler


class Loader(cached.Loader):
    is_usable = True

    def include_pug_sources(self, contents):
        """Lets fetch top level pug includes to enable  mixins"""
        match = re.search(r'^include (.*)$', contents, re.MULTILINE)
        while match:
            mixin_name = match.groups()[0]
            origin = [o for o in self.get_template_sources(mixin_name)][0]
            template = origin.loader.get_contents(origin)
            template = self.include_pug_sources(template)
            contents = re.sub(r'^include (.*)$', template, contents, flags=re.MULTILINE)
            match = re.search(r'^include (.*)$', contents, re.MULTILINE)
        return contents

    def get_contents(self, origin):
        contents = origin.loader.get_contents(origin)
        if os.path.splitext(origin.template_name)[1] in ('.pug', '.jade'):
            contents = self.include_pug_sources(contents)
            contents = process(
                contents, filename=origin.template_name, compiler=Compiler
            )
        return contents

    def get_template(self, template_name, **kwargs):
        """
        Uses cache if debug is False, otherwise re-reads from file system.
        """
        if getattr(settings, 'TEMPLATE_DEBUG', settings.DEBUG):
            try:
                return super(cached.Loader, self).get_template(template_name, **kwargs)
            # TODO: Change IOError to FileNotFoundError after future==0.17.0
            except IOError:
                raise TemplateDoesNotExist(template_name)

        return super(Loader, self).get_template(template_name, **kwargs)
